import hashlib
import itertools
from sqlalchemy import (
    Table,
    Column,
    Integer,
    Unicode,
    UnicodeText,
    DateTime,
    String,
    ForeignKey,
    )

from sqlalchemy.sql import (
    func,
    and_,
    )

from sqlalchemy.orm import (
    scoped_session,
    sessionmaker,
    relationship,
    backref,
    column_property,
    )

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method

DBSession = scoped_session(sessionmaker())
Base = declarative_base()

class TracerMixin(object):
    created = Column(DateTime, default=func.current_timestamp())
    updated = Column(DateTime, default=func.current_timestamp(),
                     onupdate=func.current_timestamp())

permission_group = Table(
    "permission_group",
    Base.metadata,
    Column('permission_id', Integer, ForeignKey('permissions.id')),
    Column('group_id', Integer, ForeignKey('groups.id')))

group_user = Table(
    "group_user",
    Base.metadata,
    Column('user_id', Integer, ForeignKey('users.id')),
    Column('group_id', Integer, ForeignKey('groups.id')))

class Permission(Base, TracerMixin):
    __tablename__ = 'permissions'
    id = Column(Integer, primary_key=True)
    permission_name = Column(Unicode(255), unique=True)
    
    def __repr__(self):
        return str(self)

    def __str__(self):
        return "Permission(permission_name={permission_name})".format(
            permission_name=self.permission_name)

class Group(Base, TracerMixin):
    __tablename__ = 'groups'
    id = Column(Integer, primary_key=True)
    group_name = Column(Unicode(255), unique=True)
    permissions = relationship('Permission',
                               secondary=permission_group,
                               backref="groups")
    permission_names = association_proxy('permissions',
                                         'permission_name')

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "Group(group_name={group_name})".format(
            group_name=self.group_name)

class User(Base, TracerMixin):
    __tablename__ = 'users'
    query = DBSession.query_property()
    id = Column(Integer, primary_key=True)
    user_name = Column(Unicode(255), unique=True)
    first_name = Column(Unicode(255))
    last_name = Column(Unicode(255))
    full_name = column_property(first_name + last_name)
    _password = Column("password", String(255))
    groups = relationship('Group',
                          secondary=group_user,
                          backref='users')

    @hybrid_property
    def permissions(self):
        return list(itertools.chain(*[g.permissions for g in self.groups]))

    def _set_password(self, password):
        self._password = hashlib.sha1(password).hexdigest()

    password = property(fset=_set_password)

    def validate_password(self, password):
        return self._password == hashlib.sha1(password).hexdigest()

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "User(id={id}, user_name={user_name})".format(
            id=self.id,
            user_name=self.user_name.encode('utf-8'))


    @hybrid_method
    def in_group(self, group_name):
        return group_name in [g.group_name for g in self.groups]

    @in_group.expression
    def in_group(cls, group_name):
        return and_(Group.group_name==group_name,
                    Group.id==group_user.c.group_id,
                    group_user.c.user_id==cls.id)

    @hybrid_method
    def has_permission(self, permission_name):
        return permission_name in [p.permission_name for p in self.permissions]

    @has_permission.expression
    def has_permission(cls, permission_name):
        return and_(Permission.permission_name==permission_name,
                    Permission.id==permission_group.c.permission_id,
                    permission_group.c.group_id==Group.id,
                    Group.id==group_user.c.group_id,
                    group_user.c.user_id==cls.id)
        
